/*
	May 22,2019 - Starting Keyer Test
				  by Nick Kennedy, WA5BDU
				  
	The purpose of the program is to "remote control" a keyer and measure
	timings and logic to see what goes on inside.
	
	May 25 - had a lot of trouble getting it to key my Winkeyer 10. I need
	to change my dot and dash lines from HIGH and LOW output to LOW output
	for Key Down and go to INPUT status for HIGH.
	
	Some open circuit paddle voltages:
	
	Keyer     		Volts
	K3         		5.0 V
	WA5BDU	   		5.0 V    Refers to my Arduino keyer
	Super CMOS		4.5 V
	FT-991A			3.3 V
	Winkeyer 10		5.0 V    Powered by 9 V battery
*/

	#define LEDpin 13 // Pin 13 - on board LED
	//#define spkrpin 7 // for sidetone
	#define KeyOut	8
	#define DotSW 5 // Line that simulates dot paddle contact
	#define DashSW 6 // Line that simulates dash paddle contact
	#define reportM 7 // ground at startup to enter Morse reporting mode
	#define KeyClosed LOW // Status of paddle when actuated
	#define KeyOpen HIGH // Paddle contact not actuated
	
	
	uint16_t ditlen = 60; // initialize to 20 WPM values
	uint16_t dahlen = 180;
	uint16_t spacetime = 60;
	uint8_t keydntime; // for figuring ditlen & dahlen
	uint8_t firstDotms;
	
	int idleStatus; // Status of TX line with key up
	uint8_t timeToClose; // ms between paddle close and TX line change
	uint8_t misc_count = 0;
	
	bool ModeA = false; // true for Mode A, false for Mode B
	bool modeBdot; // dash insertion works from string of dots
	bool modeBdash; // dot insertion works from string of dahs
	bool dotClosed = false;
	bool dashClosed = false;
	unsigned long timing_start;
	float WPM;
	float dotR, dashR;
	uint8_t paddlehold = 0; // time before new ele not generated
	
	bool ultimatic; // true if ultimatic mode
	bool uFromDot; // true if ultimatic from dot closed first
	bool uFromDash; // true if ultimatic from dash closed first
	bool isIambic;
	
	bool reportMorse = false; // true if test reporting by Morse desired
	
	String SerialMsg = ""; // User will input string to be sent by keyer
	char numBuffer[] = "    "; // hold for conversion int to ascii
	                           // big enough for nn.n
	
	
	

// **********************************************************************
//                                                                      *
//                                S E T U P                             *
//                                                                      *
// **********************************************************************
  	
	void setup()
	{
		pinMode(DotSW, INPUT);
		pinMode(DashSW, INPUT);
		pinMode(KeyOut, INPUT_PULLUP); // This is keyer's output to TX
		pinMode(reportM, INPUT_PULLUP); // boot-up option for report by Morse
		pinMode(LEDpin, OUTPUT);
		digitalWrite(LEDpin, LOW); 
		//digitalWrite(DotSW, KeyOpen);
		//digitalWrite(DashSW, KeyOpen);
		openDot();
		openDash();
		
		Serial.begin(9600);
		Serial.println(F("\nWA5BDU Keyer Analyzer V1.0"));
		
		
		delay(200);
		
		// if the reportM pin is low on startup, use Morse report mode
		if(digitalRead(reportM) == LOW) 
		{
			reportMorse = true;
			//Serial.println("\nFYI - reportMorse is true");	
		}
		

		//SerialMsg = "A new string";
		//SerialMsg = "A bigger new string"; // is this legal?
		// No error - apparently OK
	}


   
// **********************************************************************************
//                                                                                  *
//                              M A I N   L O O P                                   *
//                                                                                  *
// **********************************************************************************
	
	void loop()
	{
		char choice; // rev 1r11 - choice was int, but now allow alpha and no CR required	
		
	if(reportMorse) autoRun(); // menu choices automatic in Morse mode
	
	openDash();
	openDot();
	
   do {
     choice = 'x'; // start needing valid choice
     Serial.println(); //send CR+LF

 //    while (Serial.available() > 0) Serial.read();  // clear the serial buffer
     Serial.println();
     Serial.println(F("Select:\r\n"));
     Serial.println(F("   0 - Verify functionality"));
     Serial.println(F("   1 - Timings, speed, ratios"));
     Serial.println(F("   2 - Check Iambic & Ultimatic operation")); 
     Serial.println(F("   3 - Check Mode A or B")); 
	 Serial.println(F("   4 - Check time for same element latch"));
	 Serial.println(F("   5 - Dot paddle open/close"));
	 Serial.println(F("   6 - Dash paddle open/close"));
	 Serial.println(F("   7 - Send PARIS for WPM"));	 
	 Serial.println(F("   * - Send serial port text"));	  

     Serial.println();
 

       Serial.print(F("Choice: "));     
 	    clear_serial_buffer();
    
		while (!Serial.available());
			choice = toupper(Serial.read()); 
	 
     switch(choice)
     {
       case '1':
            doTimings(); // 
            break;
			
        case '2':
            iambicTest(); // 
            break;
			
		case '3':
			  modeABtest(); //
			  break;

		 case '4': // 
			  extraCharTest(); // 
			  break;
			  
		 case '5': // 
			  dotOpenClose();// toggle dot switch
			  break;
			  
		 case '6': //
			  dashOpenClose();// toggle dash switch
			  break;
			  
		 case '7': //
			  sendPARIS();
			  break;			  
			  	
		 case '*': 
			  sendSerialTXT(); 
			  break;	
	
		 case '0': 
				VerifyFunc(); // Make sure dot & dash paddles work as expected
				break;
				
          default:
             break;
     }     

	 // clear_serial_buffer();
	  while (!Serial.available()); // Hold reprinting menu until user hits key
	  
   }  while (1); // main menu infinite DO loop
		
}

// ********************  END OF MAIN LOOP *********************************


// ************* AUTO RUN MENU ITEMS WHEN IN MORSE MODE *******************

// This function will call each test and allow it to report, then the 
// program will end.

	void autoRun()
	{
		VerifyFunc(); 
		delay(2000);
		doTimings();
		delay(2000);
		iambicTest();
		if(isIambic) modeABtest();
		delay(2000);
		extraCharTest();
		delay(2000);
		SerialMsg = "END OF TESTS = WA5BDU TU ";
		sendSerialTXT();
		while(1); 
		
	}	

// *************  SEND A SPECIFIED NUMBER OF DITS *************************
	
	bool send_dits(uint8_t number)
	{	
		uint8_t ditcount = 0;
		bool statusback = true; // good return status
		timing_start = millis();
		//digitalWrite(DotSW, KeyClosed); // close dot paddle
		closeDot();
		
		
		while(digitalRead(KeyOut) == idleStatus) // stay until TX line change
		{
			timeToClose = millis() - timing_start;
			if(timeToClose > 100)
			{
				statusback = false; 
				goto quitdits;
			}
		}
		
		//Serial.println("Point A");
		// Now monitoring dit and counting dits
		
		while(ditcount < number - 1)
		{
			//Serial.println(ditcount, DEC);
			while(digitalRead(KeyOut) != idleStatus); // stay while key down
			ditcount++;
			while(digitalRead(KeyOut) == idleStatus); // wait out space
		}
		
		quitdits:;
		// digitalWrite(DotSW, KeyOpen);
		openDot();
		return statusback;
		
	}
		
		

// *************  SEND A SPECIFIED NUMBER OF DAHS *************************
	
	bool send_dahs(uint8_t number)
	{	
		uint8_t dahcount = 0;
		bool statusback = true; // good return status
		timing_start = millis();
		//digitalWrite(DashSW, KeyClosed); // close dot paddle
		closeDash();
		
		
		while(digitalRead(KeyOut) == idleStatus) // stay until TX line change
		{
			timeToClose = millis() - timing_start;
			if(timeToClose > 100)
			{
				statusback = false; 
				goto quitdahs;
			}
		}
		
		//Serial.println("Point A");
		// Now monitoring dah and counting dahs
		
		while(dahcount < number - 1)
		{
			//Serial.println(ditcount, DEC);
			while(digitalRead(KeyOut) != idleStatus); // stay while key down
			dahcount++;
			while(digitalRead(KeyOut) == idleStatus); // wait out space
		}
		
		quitdahs:;
		// digitalWrite(DashSW, KeyOpen);
		openDash();
		return statusback;
		
	}	

// Send letter C - as demonstration ...
	
	void sendA_C()
	{
		dashAndHalf();
		dotAndHalf();
		dashAndHalf();
		dotAndHalf();
		delay(spacetime * 4);
	}


// ***************  Menu function for doing basic timings ***********************

	void doTimings()
	{
		sayTiming();
		if(getTimes())
		{
		Serial.print(F("\nDot length ms is "));
		Serial.print(ditlen, DEC);
		Serial.print(F("\nDash length ms is "));
		Serial.print(dahlen, DEC);		
		Serial.print(F("\nSpace length ms is "));
		Serial.println(spacetime, DEC);
		Serial.print(F("1st dot length ms is "));
		Serial.println(firstDotms, DEC);
		Serial.print(F("\nSpeed: "));
		Serial.print(WPM,1);
		Serial.print(F("\nDot Ratio: "));
		Serial.print(dotR, 1);
		Serial.print(F("\nDash Ratio: "));
		Serial.println(dashR, 1);
		Serial.print(F("\nTime to close TX line ms is "));
		Serial.println(timeToClose, DEC);
		
		// reportMorse = false; // ************* TEMPORARY *******************
		
		if(reportMorse)
		{
			// Speed WPM
			
			delay(spacetime * 10);
			SerialMsg = ("SPEED IS  ");
			dtostrf(WPM, 4, 1, numBuffer);
			SerialMsg.concat(numBuffer);
			SerialMsg.concat(" WPM =");
			sendSerialTXT();
			delay(spacetime * 8);			
			
			// announce dot time in Morse
			delay(spacetime * 15);
			SerialMsg = ("DOT MS ");
			itoa(ditlen, numBuffer, 10);
			SerialMsg.concat(numBuffer);
			sendSerialTXT();
			delay(spacetime * 8);
			
			// Now dash time
			
			delay(spacetime * 10);
			SerialMsg = ("DASH MS ");
			itoa(dahlen, numBuffer, 10);
			SerialMsg.concat(numBuffer);
			sendSerialTXT();
			delay(spacetime * 8);

			// Now space time
			
			delay(spacetime * 10);
			SerialMsg = ("SPACE MS ");
			itoa(spacetime, numBuffer, 10);
			SerialMsg.concat(numBuffer);
			SerialMsg.concat(" =");
			sendSerialTXT();
			delay(spacetime * 8);

			// Now first dot time
			
			delay(spacetime * 10);
			SerialMsg = ("1ST DOT MS ");
			itoa(firstDotms, numBuffer, 10);
			SerialMsg.concat(numBuffer);
			sendSerialTXT();
			delay(spacetime * 8);
			
			// Dot ratio
			
			delay(spacetime * 10);
			SerialMsg = ("DOT RATIO  ");
			dtostrf(dotR, 3, 1, numBuffer);
			SerialMsg.concat(numBuffer);
			sendSerialTXT();
			delay(spacetime * 8);
			
			// Dash ratio
			
			delay(spacetime * 10);
			SerialMsg = ("DASH RATIO  ");
			dtostrf(dashR, 3, 1, numBuffer);
			SerialMsg.concat(numBuffer);
			SerialMsg.concat(" =");
			sendSerialTXT();
			delay(spacetime * 8);
			
			// Time to close TX line
			
			delay(spacetime * 10);
			SerialMsg = ("MS TO TX ");
			itoa(timeToClose, numBuffer, 10);
			SerialMsg.concat(numBuffer);
			sendSerialTXT();
			delay(spacetime * 8);		
			
		}
		
		}
		else 
		{
			if(reportMorse)
			{
				delay(1000);
				SerialMsg = "ERROR ";
				sendSerialTXT();
			}
			Serial.println(F("\n *** Keyed line not changing state"));
			openDash();
			openDot();
			Serial.println(F(" *** Program is suspended. Reset required."));
			while(1);
		}
		//reportMorse = true; // **************** TEMPORARY **********************
		
	}

// ************  GET IDLE KEYED LINE STATUS HIGH OR LOW *******************
// *************   AND REPORT *********************************************

		void getIdleStatus()
		{
			idleStatus = digitalRead(KeyOut); // Assume key-up, get status of that
			
			//Note: I found by experiment that digitalRead(KeyOut) returns 0 
			// when the keyed line is LOW and 1 when it is HIGH. I convert to 
			// LOW and HIGH nomenclature below.
			
			Serial.print(F("\nKey Out idle state: "));
			if(idleStatus) Serial.print(F("HIGH"));
			else Serial.print(F("LOW"));
			
			if(reportMorse)
			{ 
				SerialMsg = "IDLE STATE IS ";
				if(idleStatus) SerialMsg.concat("HIGH");
			else SerialMsg.concat("LOW = ");
				sendSerialTXT();
				delay(500);
			}
		}

// ***********  START TIMING ON millis() change ***************************
// *** Note: Can take up to a full millisecond ****************************

	void millisSynch()
	{
		unsigned long timenow;
	
		timenow = millis();
		while(timenow == millis());
		// returns when millis changes
	}
/* ************************************************************************
	
 **************  GET DOT, DASH AND SPACE TIMES, WPM, RATIOS ***************

 **************************************************************************/
 
	
 
	bool getTimes()
	{
		
		// Get idle status of key out line and report:
		
		getIdleStatus();
		
		bool TXlinechanged = true;

// I'm adding a section to measure the first dot, since some keyers and rigs
// have been reported to shorten the first one.

	
		// first hold dot paddle closed ~4 dots and also
		// measure time for TX line (keyed line) to close after
		// paddle closes. Then I measure 1st dot.
		
		millisSynch();
		closeDot();
		timing_start = millis(); // getting time to key down
		waitForKeyDNi();
		timeToClose = millis() - timing_start;
		timing_start = millis(); // re-start timing to get 1st dot time
		waitForKeyUPi();
		firstDotms = millis() - timing_start;
		
		delay(ditlen * 6);
		
		// Next get into space time frame
		
		while(getTXstatus() != idleStatus); // Wait for NOT key down
		
		// Now, keyed line status has just gone UP
		
		timing_start = millis(); // to check for unresponsive keyed line
		
		while(getTXstatus() == idleStatus) // waiting for key down
		{
			if((millis() - timing_start) > 255)  // taking too long
			{
				TXlinechanged = false; 
				goto dropout;
			}
		}
			
		
		timing_start = millis(); //Start timing the dot
		
		while(getTXstatus() != idleStatus); // Wait for key UP again
		
		ditlen = millis() - timing_start; // dot has ended - record time
		
		
		timing_start = millis(); // Now we measure space time
		
		while(getTXstatus() == idleStatus); // waiting for key down	

		spacetime = millis() - timing_start;
		
		//  ******* Now measure dash time *******
		
		openDot();
		
		closeDash();
		
		delay(dahlen * 3 + ditlen * 3); // Hold dash paddle closed ~ 3 dash
		
		
		
		while(getTXstatus() != idleStatus); // Wait for NOT key down
		
		// Next wait for keyed line to go to 'UP' status. Getting rid of
		// any leftover dot
		
		while(getTXstatus() == idleStatus); // waiting for key down
		
		timing_start = millis(); // Start timing the dash
		
		while(getTXstatus() != idleStatus); // Wait for key UP again
		
		dahlen = millis() - timing_start;	
				
		
		openDash();
		
		// Below, speed is 1200 / ditlen when the ratio is perfect 1:1:3, but 
		// I'm doing an averaging of dot, dash and space length for possible
		// better calculation when the ratios aren't perfect.
		
		WPM = 1200.0 / ( ((float) ditlen + ((float) dahlen)/3.0 + (float) spacetime)/3.0);
		
		dotR = (float) ditlen / (float) spacetime;
		dashR = (float) dahlen / (float) spacetime;
		
		dropout:;
		
		return TXlinechanged;
		
	}
		
	void sayTiming()
	{
		Serial.println(F("\nThe program will send a string of dots, then dashes\n"
		                "to measure timings, speed and ratios."));
	}
	

// *************** WAIT FOR KEY DOWN **************************************

// First waits for key up status, then waits for key to close again
// So finds the start of the next full Key DOWN period

	void waitForKeyDN()
	{
		while(getTXstatus() != idleStatus); // Wait for NOT key down
		
		// Next wait for keyed line to go to 'UP' status
		
		while(getTXstatus() == idleStatus); // waiting for key down
	}

// waitForKeyUP waits for key DOWN first, then returns when key UP
// is detected. So finds the start of the next full Key UP period
	
	void waitForKeyUP() // waits for the START OF a key up period
	{
		
		while(getTXstatus() == idleStatus); // start condition is Key UP
		
		while(getTXstatus() != idleStatus); // return when key returns to UP
	}

// This one returns as soon as key DOWN is detected. 'i' means 'immediate"

	void waitForKeyDNi()
	{
		
		while(getTXstatus() == idleStatus); // waiting for key down
	}

	void waitForKeyUPi() // this one returns as soon as key UP detected
	{
		
		//while(getTXstatus() == idleStatus); // start condition is Key UP
		
		while(getTXstatus() != idleStatus); // return when key returns to UP
	}		
		
	
// *************  CLOSE AND OPEN DOT AND DASH SWITCHES ********************

	void dotOpenClose() // Toggle open and closed states
	{
		if(dotClosed) 
		{
		openDot();
		Serial.println(F("\n Dot paddle is Open"));
		}
		else
		{
		closeDot();
		Serial.println(F("\n Dot paddle is Closed"));
		}
		
	}

	void closeDot()
	{
		pinMode(DotSW, OUTPUT);
		digitalWrite(DotSW, LOW);
		//digitalWrite(DotSW, KeyClosed);
		dotClosed = true;
	}
	
	void dashOpenClose()
	{
		if(dashClosed) 
		{
		openDash();
		Serial.println(F("\n Dash paddle is Open"));
		}
		else
		{
		closeDash();
		Serial.println(F("\n Dash paddle is Closed"));
		}
		
	}		
	void openDot()
	{
		pinMode(DotSW, INPUT);
		// digitalWrite(DotSW, KeyOpen);
		dotClosed = false;
	}


	void closeDash()
	{
		pinMode(DashSW, OUTPUT);
		digitalWrite(DashSW, LOW);		
		//digitalWrite(DashSW, KeyClosed);
		dashClosed = true;
	}
	

	void openDash()
	{
		pinMode(DashSW, INPUT);
		// digitalWrite(DashSW, KeyOpen);
		dashClosed = false;
	}

// **************** GET KEYER OUTPUT LINE STATUS **************************

// Note that you write HIGH or LOW, you Read type INT

	int getTXstatus()
	{
		return digitalRead(KeyOut);
	}
	
// ************** DOT AND A HALF ******************************************

	void dotAndHalf()
	{
		closeDot();
		waitForKeyDNi();
		delay(5);
		openDot();
		while(getTXstatus() != idleStatus); // wait for end of key down
		delay(spacetime/2); // wait until half of following space to return
	}
	
// ************** DASH AND A HALF ******************************************

	void dashAndHalf()
	{
		closeDash();
		waitForKeyDNi();
		delay(5);
		openDash();
		while(getTXstatus() != idleStatus); // wait for end of key down
		delay(spacetime/2); // wait until half of following space to return
	}
	
	
// *********** MEASURE NEXT ELEMENT AND DETERMINE DOT OR DASH ****************

	bool DotOrDash()
	{
		uint16_t eletime;
		bool retflag; // true for dot, false for dash
		
		waitForKeyUPi(); // Wait for NOT key down
		
		waitForKeyDN(); // waiting for start of key down period
		
		timing_start = millis(); // Start timing the element
		
		waitForKeyUPi(); // Wait for key UP again
		
		eletime = millis() - timing_start;
		
		// my criterion is < 1.5 ditlen = dot, >= 1.5 ditlen, dash
		
		if(eletime < (ditlen + ditlen/2)) retflag = true;
		else retflag = false;
		
		return retflag;
	}

void	iambicTest()
	{

		//delay(1000);
		sayIambic();
		delay(1000);
		isIambic = CheckIambic();
		Serial.print(F("\nLogic is: "));
		if(isIambic) 
		{
			Serial.println(F("Iambic"));
			
		}

		
		else 
		{
			Serial.println(F("Non-iambic"));
			sayUltimaticTest(); // iambic failed, so do ultimatic test
			ultimaticTest(); // do the test ...
			sayUltResults(); // report the results
		}
		
		if(reportMorse)
		{	
			delay(spacetime * 15);
			SerialMsg = "LOGIC IS ";
			if(isIambic) 
			{
				SerialMsg.concat("IAMBIC =");
			}
			else
			{
				SerialMsg.concat("NOT IAMBIC =");
			}
			sendSerialTXT();
			delay(spacetime * 10);
		}
	}
		
		
// ******************  CHECK FOR IAMBIC OPERATION *************************

// Iambic means that if both paddles are closed, alternating dots and dashes
// are generated.
		
	bool CheckIambic()
	{
		bool retvalue;
		bool lasttype, savenow;
		//uint8_t savmyi;
		

		closeDot();
		closeDash();
		delay(ditlen * 15); // dah-dit is 5 ditlens
		lasttype = DotOrDash();
		
		for(uint8_t i = 0; i < 4; i++)
		{
			//savmyi = i;
			savenow = DotOrDash();
			if(savenow != lasttype)
			{
				retvalue = true; // true means iambic
				lasttype = savenow; // alternate
			}
			else 
			{
				retvalue = false; // false is non-iambic
				goto stopit;
			}
		}
		stopit:;
		openDash();
		openDot();
		//Serial.print("i = ");
		//Serial.println(savmyi, DEC);
		return retvalue;
	}
				
	void sayIambic()
	{
		Serial.println(F("\nThe program will close both paddles briefly\n"
		                "to determine iambic or non-iambic logic."));
	}		

// ************ TEST FOR ULTIMATIC MODE ***********************************

// This should be integrated into iambic test, but since that's already 
// working, I'll do a separate routine for ULT. If the iambic test
// comes out false, I'll continue by calling this test for ultimatic.

// The test will work like this:  
// Close dot paddle for 0.5 seconds to start a string of dots. Then close
// the dash paddle. See if two consecutive dashes occur. If so, it's
// ultimatic from the dot side. Then do the same thing from the dash
// side to verify both sides.

	void ultimaticTest()
	{
		uFromDot = false; 
		uFromDash = false;
		ultimatic = false;
		
		delay(1000);
		closeDot(); // start string of dots
		delay(ditlen * 8); // about 4 dits
		closeDash(); // then close dash paddle
		delay(ditlen + ditlen/2); // make sure dash closed long enough
		if(DotOrDash()) goto faildot; // found dot - not iambic
		if(!DotOrDash()) uFromDot = true; // 2nd dash makes it true
		
		faildot:;
		
		openDash();
		openDot();	

		delay(500);
		
		 // Now test from dash side closed first		
		closeDash(); // initiate string of dashes
		delay(dahlen * 4 + spacetime * 4 ); // about four dashes		

		closeDot(); // then close dot paddle
		delay(dahlen + dahlen/3); // make sure dot closed long enough
		if(!DotOrDash()) goto faildash; // found dash - not iambic
		if(DotOrDash()) uFromDash = true; // 2nd dash makes it true
		
		faildash:;
		
		openDash();
		openDot();

		if(uFromDash && uFromDot) ultimatic = true; // both sides true = true

	}		

	void sayUltimaticTest()
	{
		Serial.println(F("Now testing for Ultimatic logic ... "));
	}
	
	void sayUltResults()
	{
		if(ultimatic)
		{
			Serial.println(F("\nKeyer has Ultimatic logic."));
		}
		else // below instances shouldn't really ever happen ...
		{
			if(uFromDot)
			{
				Serial.println(F("\nUltimatic from dot side only."));
			}
			if(uFromDash)
			{
				Serial.println(F("\nUltimatic from dash side only."));
			}
		}
		
		if(reportMorse)
		{
			delay(spacetime * 15);
			SerialMsg = "LOGIC IS ";
			if(ultimatic) SerialMsg.concat("ULTIMATIC");
			else SerialMsg.concat("NOT ULTIMATIC");
			SerialMsg.concat(" =");
			sendSerialTXT();
			delay(spacetime * 10);
		}
	}
	
// ******************  TEST FOR MODE A VERSUS MODE B LOGIC ****************

// I want to plan for contingency of it working on a string of dits but
// not on dahs, or vice-versa, which leads me to a "sort of" Mode B

// So I'll make two more bools modeBdot true when it works in string of 
// dots and modeBdash when it works in string of dashes. If both are true
// it's fully Mode B


	void checkModeAB()
	{
		modeBdash = false; // true if I get dot insertion from dash string
		modeBdot = false; // true if I get dash insertion from dot string
		
		closeDot(); // start a string of ~4 dits
		delay(ditlen * 8);
		waitForKeyDN();
		delay(ditlen/3);// close paddle 1/3rd of the way in
		closeDash();
		delay(ditlen/3); //keep it closed next 1/3rd
		openDash();
		
		// Now we check whether next element is a dash ...
		
		if(DotOrDash()) // true if DOT was detected
		{
			ModeA = true; // no dash inserted so Mode A
			modeBdot = false;
		}
		else 
		{
			ModeA = false; // looks like Mode B
			modeBdot = true; 
		}
		
		openDot(); //
		delay(500);
	
	
// ********** Now test in string of dahs **********************************
		
		
		closeDash(); // start a string of ~3 dahs
		delay(ditlen * 12);
		waitForKeyDN();
		delay(dahlen/3);// close paddle 1/3rd of the way in
		closeDot();
		delay(dahlen/3); //keep it closed next 1/3rd
		openDot();
		
		// Now we check whether next element is a dot ...
		
		if(DotOrDash()) 
		{
			ModeA = false; // true dot inserted so Mode B
			modeBdash = true;
		}
		else 
		{
			ModeA = true;
			modeBdash = false;
		}
		
		openDash(); //
		delay(500);		
		
	}
	
	void sayModeAB()
	{
	Serial.println(F("\nNow test for dot and dash insertion. Mode A or B"));
	
	}

// *** The menu calls this routine to announce the ModeA/B test, run the
//     test, and report the results ***

	void modeABtest()
	{
		delay(1000);
		sayModeAB();
		checkModeAB();
		delay(spacetime * 15);
		if(reportMorse) SerialMsg = "= MODE ";
		
		if(modeBdash && !modeBdot)
		{
			Serial.println(F("Indeterminate: Dot insertion only."));
			goto indeterminate;
		}
		if(!modeBdash && modeBdot)
		{
			Serial.println(F("Indeterminate: Dash insertion only."));
			goto indeterminate;
		}		
		
		Serial.print(F("\nKeyer logic is Mode "));
		if(ModeA) Serial.println("A");
			
		else
		{
			Serial.println("B");
		}
		
		if(reportMorse)
		{
			delay(spacetime * 15);
			if(ModeA) SerialMsg.concat("A ");
			else SerialMsg.concat("B ");
			SerialMsg.concat(" =");
			sendSerialTXT();
			delay(spacetime * 15);
		}
		indeterminate:;
	}

// *** Menu choice to test paddle release time calls here ****
	
	void extraCharTest()
	{
		Serial.println(F("\nGetting time to release paddle before another same char is latched."));
		Serial.println(F("\n... this can take up to 15 seconds at 20 WPM"));
		
		extraChar(); // do test
		sayextraChar(); // Report results
		
	}
	
	
// ********* FIND TIME TO RELEASE PADDLE BEFORE ANOTHER CHAR SENT *********

	uint8_t extraChar()
	{

	bool NewEle = false; // true means another ele was started
	bool secondPass = false; // on 2nd pass, fine tune to 1 ms
	uint8_t charcount;
	paddlehold = 0;
	
			// I want to send 3 dits, then measure time to 4th
			
		while(!NewEle)
		{
			if(!secondPass)
			{
			paddlehold += 3; // bump up 3 ms per test initially
			}
			else
			{
				paddlehold += 1; // after 1st finding, back up & do 1 ms
			}
			closeDot();
			delay(ditlen/2);
			charcount = 1; 
			while(charcount != 3)
			{
				waitForKeyDN(); // will return when next is under way
				charcount++;
			}
			
			// Now last is under way. Wait for completion, then vary time
			// until paddle is opened until additional dit is produced
			
			waitForKeyUP();
			delay(paddlehold);
			openDot();
			
			// Next I'll check in the middle of where another element should
			// be to determine if one has been generated. If not, I extend my
			// time and go again.
			
			delay(ditlen + ditlen/2 - paddlehold); // should put in middle
			if(digitalRead(KeyOut) == idleStatus) NewEle = false;
			else NewEle = true;
			delay(spacetime * 2);
			if(paddlehold > spacetime + 5) goto quittest; // should not happen
			
			if(NewEle && !secondPass) // back up and approach 1 ms per pass
			                          // for greatest accuracy
			{
				secondPass = true;
				paddlehold = paddlehold - 5;
				NewEle = false;
			}
		}
		
		quittest:;
		
		return paddlehold;
			
	}
	
	void sayextraChar()
	{
		float percentspace;
		Serial.print(F("\nAt WPM:  "));
		Serial.println(WPM,1);
		
		Serial.println(F("\nTime to get off paddle is "));
		Serial.println(paddlehold, DEC);
		Serial.print(F("This is "));
		percentspace = ((float) paddlehold / (float) spacetime) * 100.0;
		Serial.print(percentspace, 0);
		Serial.println(F("% of a space."));
		
		if(reportMorse)
		{
			delay(spacetime * 15);
			SerialMsg = "TIME TO RELEASE ";
			dtostrf(percentspace, 3, 0, numBuffer);
			SerialMsg.concat(numBuffer);
			SerialMsg.concat(" PCT OF SPACE =");
			sendSerialTXT();
			delay(spacetime * 8);
		}
	}

// Below is to test my measured time required to get off paddle
// before same element seals in.

	void testDotRpt(uint8_t holdoff)
	{
	
		uint8_t cntdots = 0;
		
		while(cntdots < 10)
		{
			closeDot();
			delay(ditlen/2);
			waitForKeyUP();
			delay(holdoff);
			openDot();
			
			delay(spacetime + ditlen/2 - holdoff);// middle of dot, if any
			
			if(digitalRead(KeyOut) == idleStatus) cntdots = 12;
			else cntdots++;
			
		}
	}

// ***************  Verify basic functions of test hook-up*****************

/*  This routine is tricky for reporting via Morse, since I'm trying to
    see if the keyer is even functional. So I'm going to report via the
	on-board Arduino LED. I'll report up to three codes:
	
	1 flash = no DOT
	2 flash = no DASH
	3 flash = bad ratio (can't tell dot from dash)
	
	Also since I call getIdleStatus and don't know if I can send Morse, I
	temporarily set the Morse flag to false before calling it.
	
	But if everything checks out OK, there is a report of OK sent by Morse.
	
	The function tries the dot input. If it fails, it still checks the dash
	input using a default ditlen value. If it fails, there's no reason to 
	do the ratio calc.
	
	In the Morse reporting mode, the function stays in the flashing LED 
	routine forever and the user has to restart the Arduino after fixing the 
	problem(s).

	   
*/

	void VerifyFunc()
	{
		uint16_t ditlenLocal = 60; // set to default in case timing fails
		bool tempMflag;
		bool badDot = false;
		bool badDash = false;
		bool badRatio = false;
		
		tempMflag = reportMorse;
		
		if(reportMorse) reportMorse = !reportMorse;
		
		getIdleStatus();
		
		reportMorse = tempMflag; // restore Morse flag
		
		// Now borrow from my timings routine to get dot time. I need it
		// so I can tell that a dash is a dash in the next test
		// I'm not assuming Menu #1 has been performed here, so I don't
		// know the times.
		
		closeDot();
		timing_start = millis(); // this is for my initial wait for key close
		
		while(getTXstatus() == idleStatus) // waiting for key down
		{
			if((millis() - timing_start) > 255)  // taking too long
			{
				Serial.println(F("\nNo response from dot input closed."));
				openDot();
				badDot = true;
				goto tryDash;
			}
		}
			
		
		timing_start = millis(); //Start timing the dot
		delay(5); // don't release paddle too quickly
		openDot(); // but assume self-completing
		
		while(getTXstatus() != idleStatus); // Wait for key UP again
		
		ditlenLocal = millis() - timing_start; // dot has ended - record time
		
	// Now test the dash input and verify produces dash
		
		tryDash:;
		
		delay(500);

		closeDash();
		timing_start = millis();
		
		while(getTXstatus() == idleStatus) // waiting for key down
		{
			if((millis() - timing_start) > 255)  // taking too long
			{
				Serial.println(F("\nNo response from dash input closed."));
				openDash();
				badDash = true;
				goto dropout2;
			}
		}
			
		
		timing_start = millis(); //Start timing the dash
		delay(5); // don't release paddle too quickly
		openDash(); // but assume self-completing
		
		while(getTXstatus() != idleStatus); // Wait for key UP again
		
 // Below I assume if dash time is not at least 1.5 time dot time, something
 // is wrong, such as dot and dash lines swapped. 

		
		if((millis() - timing_start) < (ditlenLocal + ditlenLocal/2))
		{
			Serial.println(F("\nDot/Dash ratio incorrect."));
			badRatio = true;
			goto dropout2;
		}
		
	Serial.println(F("\nDot/Dash in line out responses are normal."));
	
	
	if(reportMorse)
	{	
		delay(spacetime * 15);
		SerialMsg = "FUNCTION CHECKS OK =";
		sendSerialTXT();
	}

	dropout2:;
	
	if(reportMorse && (badDot || badDash || badRatio))
	{
		while(1)
		{
			if(badDot) flashLED(1);
			if(badDash) flashLED(2);
			if(badRatio) flashLED(3);
		}

	}
	
	}


	void flashLED(uint8_t times)
	{
		// For each "times", LED ON 250 ms, OFF 250 ms and 1 s at end
		
		for(uint8_t i = times; i > 0; i--)
		{
			digitalWrite(LEDpin, HIGH);
			delay(250);
			digitalWrite(LEDpin, LOW);
			delay(250);
		}
		delay(750);
			
	}

// ************  ROUTINES FOR SENDING MORSE FROM TEXT *********************

// Morse characters are stored like this:  The high bit represents the first
// element of a character.  0 is a dot and 1 is a dash.  The final 1 flags
// the end of this character and is not sent.  Therefore 'B' being _... is 
// stored as 0b10001000
// Sending happens by checking the high bit and sending dot or dash as
// appropriate, then left shifting the data by one bit.  Before sending though,
// if the data = 0b10000000, then stop.  Sending of this character is finished.

// There are some gaps in ASCII codes that have characters I don't need and
// so waste some RAM, but not enough to do special remapping.
// From 44 through 90 are 47 bytes and I figure 41 are moderately necessary.

// Notice below how it's possible to break lines and even have comments
// interspersed

byte morse[47] = {0b11001110, 0b10000110, 0b01010110,// ','  '-'  '.' 44-46
					0b10010100, 0b11111100, 0b01111100,// /, 0, 1  47-49
					0b00111100, 0b00011100, 0b00001100, // 2, 3, 4 50-52
					0b00000100, 0b10000100, 0b11000100, // 5, 6, 7 53-55
					0b11100100, 0b11110100, 0b10110100, // 8, 9, : (KN) 56-58
					0b10101010, 0b01010100, 0b10001100, // ; (KR), < (AR), = (BT) 59-61
					0b00010110, 0b00110010, 0b01101010, // > (SK), ?, @ (AC) 62-64
					0b01100000, 0b10001000, 0b10101000, // ABC starts at 65 DEC
					0b10010000, 0b01000000, 0b00101000, // DEF
					0b11010000, 0b00001000, 0b00100000, // GHI
					0b01111000, 0b10110000, 0b01001000, // JKL
					0b11100000, 0b10100000, 0b11110000, // MNO
					0b01101000, 0b11011000, 0b01010000, // PQR
					0b00010000, 0b11000000, 0b00110000, // STU
					0b00011000, 0b01110000, 0b10011000, // VWX
					0b10111000, 0b11001000};             // YZ ends at 90 DEC



// ******  SEND CHARACTER STRING IN MORSE *********************************

// Using dashAndHalf and dotAndHalf, the keyer takes care of enforcing the
// other half of a space. But at the end of a character, the program must 
// add 2.5 more spacetimes for charactor space or 6.5 more for word space.

void send_code(int pointtocode)
{
	byte pattern;
	
	if(pointtocode == ' ') goto dospace;
	
	if((pointtocode < 44) || (pointtocode > 90)) // limit to what's in table
	{
		goto outOfRange; 
	}
	
	dospace:;
	
	if (pointtocode == ' ')
	{
	   word_space();
	}
	else
	{
	pointtocode -= 44; // shift ASCII pointer to begin at 0 for comma
	pattern = morse[pointtocode];
	
	while (pattern != 128)
	{
		if (pattern & 0b10000000)
		{
			dashAndHalf(); // sends a dash and delays 1/2 space
		}
		else
		{
		   dotAndHalf(); // sends a dot and delays 1/2 space
		} 
		
		pattern <<= 1;
		
	}
	char_space();
	
	outOfRange:;
	
	}
}		

	void char_space()
		{
			delay(spacetime * 2 + spacetime / 2);
		}


	void word_space()
		{
		// a word space is 7 spaces long. 3 were sent following the character
			
			  delay(4 * spacetime);
		}

	void sendPARIS()
	{
		float tstart;
		float ptime;
		
		tstart = (float) millis();
		send_code('P');
		send_code('A');
		send_code('R');
		send_code('I');
		send_code('S');
		send_code(' ');
		ptime = (float) millis() - tstart;
		WPM = 60000.0/ptime;
		Serial.print(F("\nSpeed using PARIS: "));
		Serial.print(WPM,1);
	}		

// ********* GET A STRING FROM SERIAL PORT & SEND ON KEYER ****************

// *** In a modification, I'll use the same routine for the "report in Morse"
//     mode, skipping the parts that get the text from serial port.

	void sendSerialTXT()
	{
		uint8_t MessagePtr, XChar;
		
		if(reportMorse) goto rptM;
		
		Serial.println(F("\nInput your text to be sent ..."));
		
		clear_serial_buffer();
		
		delay(100);
		clear_serial_buffer();
		
		while(!Serial.available());
		
		SerialMsg = Serial.readString();	
		SerialMsg.toUpperCase();
		
		rptM:; // start here for report in Morse mode
		
		MessagePtr = 0;
		XChar = SerialMsg[0];
		
		do
		{	
			// allow space and % to sneak by range test
			
			if(XChar == ' ') goto allowspace;
			
			if(XChar == '%')
			{
				MessagePtr = 0; // Percent sign means repeat msg
				XChar = SerialMsg[MessagePtr];
			}
			
			allowspace:;
			
			Serial.print(char(XChar));

	// Although the user won't be watching the serial output in Morse
	// mode, during development I'd like some line breaks. So after each
	// [BT] or = sign, I'm putting one to the screen.
	
			if(reportM && (XChar == '=')) Serial.println();
			
			send_code(XChar);
			MessagePtr++;			

			XChar = SerialMsg[MessagePtr];
			
			if(!reportMorse)
			{
				if(Serial.available()) XChar = NULL; // serial input to quit
			}
			
		} while	((XChar != 10) && (XChar != 13) && (XChar != NULL));
		
	}
	
	
	
// ********** Routines imported from PHSNA, mostly serial, for menu *******

// Below "do nothing" function returns to main menu from case stmt

	//void more_menu()
	//	{ return;} // Called when user hits invalid menu choice, to repeat menu



// **********  CLEAR SERIAL BUFFER ******************************************
// Below addresses a problem I had when using a routine that did the same as this
// but without the time delay.  It occurred when the sender was in process of
// sending a string of characters.  The routine would read one character and then
// check Serial.available() and find it was 0, because the next character hadn't 
// been fully shifted in yet.  At 9600 baud, a character takes 1.04 ms to transmit.
// I'll conservatively wait 10 ms after getting a byte before checking for another
// char available.  NRK 12/10/2013

  void clear_serial_buffer()
  {
	 
     while (Serial.available() > 0) 
	 {
	   delay(10);
	   Serial.read();  // get available byte
	 }
}
	 
// Print a comma and a space with no linefeed ...

	void comma_space() 
	{
	   Serial.print(", ");
	}
// Print a backspace character

	void backspace()
	{
		Serial.print("\b");
	}

	void print_8_stars()
	{
	  Serial.println(F("********"));
	 }
	 
	 void say_enter_freq()
	 {
	   Serial.println(F("Enter frequency in Hz: "));
	 }
// ******************  GET A SINGLE CHARACTER FROM SERIAL PORT **********************

	char get_char()
	{
	   clear_serial_buffer(); // clear out old trash
	   while (!Serial.available()); // wait until a char is ready
	   return (Serial.read());
	
	}		
		